# Generated by roxygen2: do not edit by hand

S3method(lrtest,twoPhaseSPML)
S3method(print,lrtest.twoPhaseSPML)
S3method(print,twoPhaseSPML)
export(DataGeneration_TPD)
export(lrtest)
export(optimTP.GA)
export(optimTP.LM)
export(twoPhase)
export(twoPhaseDesign)
export(twoPhaseHeuristic)
export(twoPhaseSPML)
export(twoPhaseSPML.control)
import(data.table)
import(stats)
importFrom(MASS,ginv)
importFrom(Matrix,colSums)
importFrom(Matrix,crossprod)
importFrom(Matrix,fac2sparse)
importFrom(Matrix,t)
importFrom(dfoptim,hjkb)
importFrom(enrichwith,enrich)
importFrom(kofnGA,kofnGA)
importFrom(nloptr,cobyla)
importFrom(parallel,clusterEvalQ)
importFrom(stats,pchisq)
importFrom(utils,combn)
importFrom(utils,stack)
importFrom(utils,tail)
