% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lrtest.twoPhaseSPML.R
\name{print.lrtest.twoPhaseSPML}
\alias{print.lrtest.twoPhaseSPML}
\title{Print method for \code{lrtest.twoPhaseSPML} objects}
\usage{
\method{print}{lrtest.twoPhaseSPML}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{an object of class \code{"lrtest.twoPhaseSPML"}, as returned by
\code{\link{lrtest.twoPhaseSPML}}.}

\item{digits}{integer; number of significant digits. Defaults to
\code{max(3L, getOption("digits") - 3L)}.}

\item{\dots}{further arguments passed to \code{\link{printCoefmat}}.}
}
\value{
\code{x} is returned invisibly.
}
\description{
Print method for \code{lrtest.twoPhaseSPML} objects
}
\seealso{
\code{\link{lrtest.twoPhaseSPML}}
}
