% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twoPhaseSPMLglm_.R
\name{twoPhaseSPML.control}
\alias{twoPhaseSPML.control}
\title{Auxiliary for Controlling the EM algorithm in twoPhaseSPML.}
\usage{
twoPhaseSPML.control(EMmaxit = 1000, EMtol = 1e-05, trace = FALSE)
}
\arguments{
\item{EMmaxit}{integer giving the maximal number of iterations of the EM algorithm.}

\item{EMtol}{numeric positive convergence tolerance for EM algorithm covnergence.}

\item{trace}{logical \code{TRUE}, prints progress at each EM iteration. Defaults to \code{FALSE}.}
}
\value{
A list with components named as the arguments.
}
\description{
Auxiliary for Controlling the EM algorithm in twoPhaseSPML.
}
\details{
The function implements the semiparametric maximum likelihood (SPML)
}
